from datetime import datetime
from typing import Any
from uuid import uuid4
import orjson

from pydantic import BaseModel, Field, UUID4


def orjson_dumps(v: Any, *, default: Any):
    return orjson.dumps(v, default).decode()


class BaseMessage(BaseModel):
    id: UUID4 = Field(default_factory=uuid4)
    create_time: datetime = Field(default_factory=datetime.now)

    class Config:
        validate_all = True
        allow_mutation = False
        allow_population_by_field_name = True
        json_loads = orjson.loads
        json_dumps = orjson_dumps

    def pre_handle(self):
        ...

    def post_handle(self):
        ...



from typing import Any, ClassVar

from pydamain.domain import Command, Event, CommandHandler, EventHandlers


sync_cmd_handler_flag = False
async_cmd_handler_flag = False


def sync_cmd_handler(m: Command, **_: Any):
    global sync_cmd_handler_flag
    sync_cmd_handler_flag = True


async def async_cmd_handler(m: Command, **_: Any):
    global async_cmd_handler_flag
    async_cmd_handler_flag = True


sync_evt_handler_flag = False
async_evt_handler_flag = False


def sync_evt_handler(m: Event, **_: Any):
    global sync_evt_handler_flag
    sync_evt_handler_flag = True


async def async_evt_handler(m: Event, **_: Any):
    global async_evt_handler_flag
    async_evt_handler_flag = True


def sync_evt_raise_exception(m: Event, **_: Any):
    raise Exception()


async def async_evt_raise_exception(m: Event, **_: Any):
    raise Exception()


def issue_example_event(m: Command, **_: Any):
    ExampleEvent().issue()


class ExampleEvent(Event):

    handlers: ClassVar[EventHandlers["ExampleEvent"]] = [
        sync_evt_handler,
        async_evt_handler,
        sync_evt_raise_exception,
        async_evt_raise_exception,
    ]


class TestCommand:
    async def test_sync_handler_handle(self):
        class ExampleCommand(Command):
            handler: ClassVar[
                CommandHandler["ExampleCommand"]
            ] = sync_cmd_handler

        await ExampleCommand().handle({})
        assert sync_cmd_handler_flag == True

    async def test_async_handler_handle(self):
        class ExampleCommand(Command):
            handler: ClassVar[
                CommandHandler["ExampleCommand"]
            ] = async_cmd_handler

        await ExampleCommand().handle({})
        assert async_cmd_handler_flag == True

    async def test_handle_events_and_ignore_exception(self):
        class ExampleCommand(Command):
            handler: ClassVar[
                CommandHandler["ExampleCommand"]
            ] = issue_example_event

        await ExampleCommand().handle({})
        assert sync_evt_handler_flag == True
        assert async_evt_handler_flag == True



import orjson
import typing

from pydantic import BaseModel


def orjson_dumps(v: typing.Any, *, default: typing.Any):
    return orjson.dumps(v, default).decode()


class BaseDomainModel(BaseModel):
    class Config(BaseModel.Config):
        validate_all = True
        allow_population_by_field_name = True
        validate_assignment = True
        orm_mode = True
        json_loads = orjson.loads
        json_dumps = orjson_dumps



from uuid import uuid4
import pytest

from pydamain.domain import Entity


class ExampleEntity(Entity):
    text: str


class TestEntity:
    def test_id_immutable(self):
        with pytest.raises(TypeError):
            ee = ExampleEntity(text="first")
            ee.id = uuid4()

    def test_id_equality(self):
        same_id = uuid4()
        first = ExampleEntity(id=same_id, text="first")
        second = ExampleEntity(id=same_id, text="second")
        assert first == second

    def test_id_hashing(self):
        same_id = uuid4()
        first = ExampleEntity(id=same_id, text="first")
        second = ExampleEntity(id=same_id, text="second")
        assert len(set([first, second])) == 1

    def test_jsonable(self):
        assert ExampleEntity(text="first").json() == r'{"text":"first"}'



from dataclasses import dataclass
from datetime import datetime, date, time
from typing import Any, TypedDict
from uuid import UUID, uuid4
import pytest

from pydantic import ValidationError

from pydamain.domain import Field
from pydamain.domain.model.base import BaseDomainModel


class ValidationDefault(BaseDomainModel):

    test: int = Field(default="exception")


class ValidationDefaultFactory(BaseDomainModel):

    test: int = Field(default_factory=lambda: "exception")


class UseableAlias(BaseDomainModel):

    id: str = Field(alias="identity")


class OrmModeObject:
    def __init__(self, id: int):
        self.id = id


class OrmModeDomainModel(BaseDomainModel):

    id: int


@dataclass
class ExampleDataClass:
    value: str


class ExampleTypedDict(TypedDict):
    value: str


class NestedBaseDomainModel(BaseDomainModel):
    value: str


class JsonableTypes(BaseDomainModel):

    none_: None

    int_: int
    float_: float
    bool_: bool

    str_: str
    list_: list[Any]
    tuple_: tuple[Any, ...]
    set_: set[Any]
    frozenset_: frozenset[Any]

    dict_: dict[Any, Any]
    typeddict_: ExampleTypedDict

    datetime_: datetime
    date_: date
    time_: time

    uuid_: UUID

    dataclass_: ExampleDataClass

    nested: NestedBaseDomainModel


class TestBaseDomainModel:
    def test_validate_default_value(self):
        with pytest.raises(ValidationError):
            ValidationDefault()

    def test_validate_default_factory_value(self):
        with pytest.raises(ValidationError):
            ValidationDefaultFactory()

    def test_useable_alias(self):
        UseableAlias(identity="other")

    def test_on_orm_mode(self):
        omo = OrmModeObject(123)
        OrmModeDomainModel.from_orm(omo)

    def test_json_translation(self):
        origin = JsonableTypes(
            none_=None,
            int_=0,
            float_=0,
            bool_=True,
            str_="",
            list_=[],
            tuple_=tuple(),
            set_=set(),
            frozenset_=frozenset(),
            dict_={},
            typeddict_=ExampleTypedDict(value=""),
            datetime_=datetime.now(),
            date_=datetime.now().date(),
            time_=datetime.now().time(),
            uuid_=uuid4(),
            dataclass_=ExampleDataClass(""),
            nested=NestedBaseDomainModel(value="nested"),
        )
        json_string = origin.json()
        parsed = JsonableTypes.parse_raw(json_string)
        assert origin == parsed



from typing import Any, ClassVar
import orjson

from aiokafka import AIOKafkaProducer  # type: ignore

from pydantic import BaseModel

from pydamain.domain import Event
from pydamain.port.out_.event_producer import AbstractEventProducer


def orjson_dumps(v: Any, *, default: Any):
    return orjson.dumps(v, default).decode()


class KafkaEvent(BaseModel):
    event_type: str
    event_data: str

    class Config:
        validate_all = True
        allow_mutation = False
        allow_population_by_field_name = True
        json_loads = orjson.loads
        json_dumps = orjson_dumps


def event_to_jsonb(event: Event):
    kafka_event = KafkaEvent(
        event_type=type(event).__name__, event_data=event.json()
    )
    return kafka_event.json().encode("utf-8")


def jsonb_to_event(jsonb: bytes):
    json = jsonb.decode("utf-8")
    kafka_event = KafkaEvent.parse_raw(json)
    event_cls = Event.__event_subclass_registry__[kafka_event.event_type]
    return event_cls.parse_raw(kafka_event.event_data)


class BaseKafkaEventProducer(AbstractEventProducer):

    host: ClassVar[str]
    port: ClassVar[int]
    topic: ClassVar[str]

    def __init__(self) -> None:
        self.aiokafka_producer = AIOKafkaProducer(
            bootstrap_servers=f"{self.host}:{self.port}",
            value_serializer=event_to_jsonb,
            compression_type="lz4",
            enable_idempotence=True,  # 멱등성
        )

    def pre_send(self, event: Event):
        ...

    def post_send(self, event: Event):
        ...

    async def send(self, event: Event, key: int | None = None):
        self.pre_send(event)
        await self.aiokafka_producer.send(  # type: ignore
            topic=self.topic, key=key, value=event
        )
        self.post_send(event)



import pytest

from pydamain.domain import ValueObject


class ExampleValueObject(ValueObject):
    text: str


class TestValueObject:
    def test_immutable(self):
        with pytest.raises(TypeError):
            evo = ExampleValueObject(text="first")
            evo.text = "second"

    def test_jsonable(self):
        assert ExampleValueObject(text="first").json() == r'{"text":"first"}'

    def test_equality(self):
        assert ExampleValueObject(text="same") == ExampleValueObject(
            text="same"
        )
        assert ExampleValueObject(text="first") != ExampleValueObject(
            text="second"
        )



from contextvars import ContextVar
from typing import Any, ClassVar, Iterable, TypeVar, cast
from weakref import WeakValueDictionary
import asyncio

from .base import BaseMessage
from .typing import AsyncHandler, Handler, SyncHandler


# ============================================================================
# Command
# ============================================================================
C = TypeVar("C", bound="Command")
CommandHandler = Handler[C, Any]


class Command(BaseMessage):

    handler: ClassVar[CommandHandler[Any]] = None  # type: ignore

    async def handle(self, deps: dict[str, Any]):
        self.pre_handle()
        events = await asyncio.create_task(self._handle(deps))
        self.post_handle()
        await handle_events(events, deps)

    async def _handle(self, deps: dict[str, Any]):
        if not self.handler:
            return list[Event]()
        with EventCatcher() as event_catcher:
            await handle(self, type(self).handler, deps)
        return event_catcher.events


# ============================================================================
# Event
# ============================================================================
E = TypeVar("E", bound="Event")
EventHandlers = Iterable[Handler[E, Any]]


class Event(BaseMessage):

    handlers: ClassVar[EventHandlers[Any]] = []

    __event_subclass_registry__: ClassVar[
        WeakValueDictionary[str, type["Event"]]
    ] = WeakValueDictionary()

    def __init_subclass__(cls) -> None:
        cls.handlers = set(cls.handlers)
        cls.__event_subclass_registry__[cls.__name__] = cls

    def issue(self):
        event_list = events_context_var.get()
        event_list.append(self)

    async def _handle(self, deps: dict[str, Any]):
        self.pre_handle()
        events = await asyncio.create_task(self.__handle(deps))
        self.post_handle()
        await handle_events(events, deps)

    async def __handle(self, deps: dict[str, Any]):
        with EventCatcher() as event_catcher:
            await asyncio.gather(
                *(handle(self, handler, deps) for handler in self.handlers),
                return_exceptions=True,
            )
        return event_catcher.events


async def handle_events(
    events: Iterable[Event],
    deps: dict[str, Any],
    ignore_exception: bool = True,
):
    coros = (event._handle(deps) for event in events)  # type: ignore
    await asyncio.gather(*coros, return_exceptions=ignore_exception)


# ============================================================================
# Handle Function
# ============================================================================
M = TypeVar("M", bound=BaseMessage)
R = TypeVar("R")


async def handle(msg: M, handler: Handler[M, R], deps: dict[str, Any]) -> R:
    if asyncio.iscoroutinefunction(handler):
        handler = cast(AsyncHandler[M, R], handler)
        return await handler(msg, **deps)
    handler = cast(SyncHandler[M, R], handler)
    return await asyncio.to_thread(handler, msg, **deps)


# ============================================================================
# Event Catcher
# ============================================================================
events_context_var: ContextVar[list[Event]] = ContextVar("events_context_var")


class EventCatcher:
    def __init__(self):
        self.events: list[Event] = []

    def __enter__(self):
        self.token = events_context_var.set(list())
        return self

    def __exit__(self, *args: tuple[Any]):
        self.events.extend(events_context_var.get())
        events_context_var.reset(self.token)



from dataclasses import dataclass
from datetime import date, datetime, time
from typing import Any, TypedDict
from uuid import UUID, uuid4
import pytest

from pydantic import ValidationError

from pydamain.domain import Field
from pydamain.domain.model.base import BaseDomainModel
from pydamain.domain.message.base import BaseMessage


@dataclass
class ExampleDataClass:
    value: str


class ExampleTypedDict(TypedDict):
    value: str


class NestedBaseDomainModel(BaseDomainModel):
    value: str


class JsonableTypes(BaseMessage):

    none_: None

    int_: int
    float_: float
    bool_: bool

    str_: str
    list_: list[Any]
    tuple_: tuple[Any, ...]
    set_: set[Any]
    frozenset_: frozenset[Any]

    datetime_: datetime
    date_: date
    time_: time

    uuid_: UUID

    dict_: dict[Any, Any]
    typeddict_: ExampleTypedDict
    dataclass_: ExampleDataClass
    nested: NestedBaseDomainModel


class ExampleBaseMessage(BaseMessage):
    value: str


class ValidationDefault(BaseMessage):
    value: str = Field(default=int)


class UseableAlias(BaseMessage):
    name: str = Field(alias="username")


class TestBaseMessage:
    def test_validate_default_value(self):
        with pytest.raises(ValidationError):
            ValidationDefault()

    def test_useable_alias(self):
        UseableAlias(username="other")

    def test_immutable(self):
        example = ExampleBaseMessage(value="immutable")
        with pytest.raises(TypeError):
            example.value = "change"

    def test_jsonable(self):
        origin = JsonableTypes(
            none_=None,
            int_=0,
            float_=0,
            bool_=True,
            str_="",
            list_=[],
            tuple_=tuple(),
            set_=set(),
            frozenset_=frozenset(),
            dict_={},
            typeddict_=ExampleTypedDict(value=""),
            datetime_=datetime.now(),
            date_=datetime.now().date(),
            time_=datetime.now().time(),
            uuid_=uuid4(),
            dataclass_=ExampleDataClass(""),
            nested=NestedBaseDomainModel(value="nested"),
        )
        json_string = origin.json()
        parsed = JsonableTypes.parse_raw(json_string)
        assert origin == parsed



from typing import Any

from pydamain.domain.message.base import BaseMessage
from pydamain.domain.message.main import handle


sync_flag = False
async_flag = False


def sync_handler(m: "ExampleMessage", **_: Any):
    global sync_flag
    sync_flag = True
    return "runned sync handler"


async def async_handler(m: "ExampleMessage", **_: Any):
    global async_flag
    async_flag = True
    return "runned async handler"


class ExampleMessage(BaseMessage):
    ...


class TestHandleMessage:
    async def test_handle_sync(self):
        m = ExampleMessage()
        result = await handle(m, sync_handler, {})
        assert sync_flag
        assert result == "runned sync handler"

    async def test_handle_async(self):
        m = ExampleMessage()
        result = await handle(m, async_handler, {})
        assert async_flag
        assert result == "runned async handler"



from typing import Any
from uuid import uuid4

from pydantic import Field, UUID4

from .base import BaseDomainModel


class ValueObject(BaseDomainModel):
    class Config(BaseDomainModel.Config):
        allow_mutation = False

    def __eq__(self, other: Any):
        if isinstance(other, type(self)):
            return self.__dict__ == other.__dict__
        return NotImplemented


class Entity(BaseDomainModel):
    id: UUID4 = Field(
        default_factory=uuid4, exclude=True, allow_mutation=False
    )

    class Config(BaseDomainModel.Config):
        ...

    def __eq__(self, other: Any):
        if isinstance(other, type(self)):
            return self.id == other.id
        return NotImplemented

    def __hash__(self):  # type: ignore
        return hash(self.id)


class Aggregate(Entity):
    class Config(Entity.Config):
        ...



from typing import Any, ClassVar

import pytest

from pydamain.domain import Event, EventHandlers


sync_flag = False
async_flag = False


def sync_handler(m: "ExEvent", **_: Any):
    global sync_flag
    sync_flag = True


async def async_handler(m: "ExEvent", **_: Any):
    global async_flag
    async_flag = True


nested_sync_flag = False
nested_async_flag = False


def nested_sync_handler(m: "ExNestedEvent", **_: Any):
    global nested_sync_flag
    nested_sync_flag = True


async def nested_async_handler(m: "ExNestedEvent", **_: Any):
    global nested_async_flag
    nested_async_flag = True


def sync_handler_raise_exception(m: "ExEvent", **_: Any):
    raise Exception()


async def async_handler_raise_exception(m: "ExEvent", **_: Any):
    raise Exception()


async def issue_nested_events(m: "ExEvent", **_: Any):
    ExNestedEvent().issue()


class ExEvent(Event):

    handlers: ClassVar[EventHandlers["ExEvent"]] = [
        sync_handler,
        async_handler,
        sync_handler_raise_exception,
        async_handler_raise_exception,
        issue_nested_events,
    ]


class ExNestedEvent(Event):

    handlers: ClassVar[EventHandlers["ExNestedEvent"]] = [
        nested_sync_handler,
        nested_async_handler,
    ]


class TestEvent:
    def test_need_event_catcher_context(self):
        e = ExEvent()
        with pytest.raises(LookupError):
            e.issue()

    async def test_handle(self):
        await ExEvent()._handle({})  # type: ignore
        assert sync_flag == True
        assert async_flag == True
        assert nested_sync_flag == True
        assert nested_async_flag == True

    @pytest.mark.skip(reason="NotImplemented")
    async def test_block_circular_event_issue(self):
        def issue_one(m: Event, **_: Any):
            OneEvent().issue()

        async def issue_two(m: Event, **_: Any):
            TwoEvent().issue()

        class OneEvent(Event):
            handlers: ClassVar[EventHandlers["OneEvent"]] = [issue_two]

        class TwoEvent(Event):
            handlers: ClassVar[EventHandlers["TwoEvent"]] = [issue_one]

        with pytest.raises(Exception):
            await OneEvent()._handle({})  # type: ignore



from typing import Any, ClassVar, cast

from aiokafka import AIOKafkaConsumer  # type: ignore

from pydantic import BaseModel

from pydamain.domain import Event
from pydamain.adapter.out_.event_producer import jsonb_to_event
from pydamain.port.in_.event_consumer import AbstractEventConsumer


class KafkaEvent(BaseModel):
    event_type: bytes
    event: bytes


class BaseKafkaEventConsumer(AbstractEventConsumer):

    host: ClassVar[str]
    port: ClassVar[int]
    topic: ClassVar[str]

    def __init__(self, deps: dict[str, Any]) -> None:
        self.aiokafka_consumer = AIOKafkaConsumer(
            self.topic,
            bootstrap_servers=f"{self.host}:{self.port}",
            value_deserializer=jsonb_to_event,
            group_id=type(self).__name__,
            enable_auto_commit=False,
            auto_offset_reset="earliest",
        )
        self.deps = deps

    def pre_consume(self, event: Event):
        ...

    def post_consume(self, event: Event):
        ...

    async def consume(self):
        async for event in self.aiokafka_consumer:  # type: ignore
            event = cast(Event, event)
            self.pre_consume(event)
            await event._handle(self.deps)  # type: ignore
            await self.aiokafka_consumer.commit()  # type: ignore
            self.post_consume(event)



from typing import Any, Awaitable, Protocol, TypeVar


T_contra = TypeVar("T_contra", contravariant=True)
T_co = TypeVar("T_co", covariant=True)


class SyncHandler(Protocol[T_contra, T_co]):
    __name__: str

    def __call__(self, __msg: T_contra, /, **__kwg: Any) -> T_co:
        ...


class AsyncHandler(Protocol[T_contra, T_co]):
    __name__: str

    async def __call__(self, __msg: T_contra, /, **__kwg: Any) -> T_co:
        ...


class Handler(Protocol[T_contra, T_co]):
    __name__: str

    def __call__(
        self, __msg: T_contra, /, **__kwg: Any
    ) -> T_co | Awaitable[T_co]:
        ...



import pytest

from pydamain.domain import Event
from pydamain.domain.message.main import EventCatcher


class TestEventCatcher:
    def test_required_event_catcher(self):
        with pytest.raises(LookupError):
            Event().issue()

    def test_catch_event_nested(self):
        with EventCatcher() as ec1:
            Event().issue()
            with EventCatcher() as ec2:
                Event().issue()
                Event().issue()
                Event().issue()
            Event().issue()
        assert len(ec1.events) == 2
        assert len(ec2.events) == 3
    
    def test_catch_event_sibling(self):
        with (
            EventCatcher() as ec1,
            EventCatcher() as ec2,
            EventCatcher() as ec3,
        ):
            Event().issue()
            Event().issue()
            Event().issue()
        assert ec1.events == []
        assert ec2.events == []
        assert len(ec3.events) == 3
